#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/signal.h>


#define BAUDRATE B9600
#define SERIALDEVICE "/dev/com2"

#define _POSIX_SOURCE 1 /* POSIX compliant source */

#define FALSE 0
#define TRUE 1

#define BLACK 0
#define WHITE 1


int from_crafty,to_crafty;
int fd;
pid_t crafty_pid;

FILE * fo;

fd_set fdread,fdwrite;
struct termios oldtio,newtio,oldstdtio,newstdtio,oldstdouttio;
struct sigaction sa;
 

void boardinit(void);
void movetoboard(char *move);
void movetocrafty(char *move);
int  excommand(char *command, char *buffer,int n);
int debug=FALSE;
int color=BLACK;

int main(int argc, char **argv) 
  {
/*
 ----------------------------------------------------------
|                                                          |
|  fork() to spawn a new process, then execve() to execute |
|  crafty instead of the cloned copy of myboard created by |
|  the fork() call.                                        |
|                                                          |
 ----------------------------------------------------------
*/
  char inbuffer[1025];
  char outbuffer[1025];
  char boardbuffer[1025];
  char command [1025];
  char posmove[10];
  ssize_t readnum;
  struct timeval tv;
  int i,j,k,l,n,lock,postpone;
  char *argv1[1]={0};
  int tp[2], fp[2];
  int ch;
  int side;

  printf("\n Maecenas 0.2 Released under GPL   (c) P.Violini \n\n"); 
  if(argc>1)
    if(argv[1][0]=='-')
      if(argv[1][1]=='d')
        debug=TRUE; 

  if(debug)
    {
    fo=fopen("maecenas.log","w");
    fprintf(fo,"Executing crafty.\n");
    }
  printf("Executing crafty.\n");

  signal(SIGPIPE, SIG_IGN);

  pipe(tp);
  pipe(fp);
  to_crafty=tp[1];
  from_crafty=fp[0];
/*
 ------------------------------------------------
|                                                |
|  child process has to force stdin, stdout and  |
|  stderr to use the two pipes we just created.  |
|                                                |
 ------------------------------------------------
*/
  if ((crafty_pid = fork()) == 0) 
    {
    dup2(tp[0], fileno(stdin));
    dup2(fp[1], fileno(stdout));
    close(tp[0]);
    close(tp[1]);
    close(fp[0]);
    close(fp[1]);
    dup2(1, fileno(stderr));
    execv("crafty",argv1);
    printf("ERROR: unable to execute crafty!\n");
    if(debug)
      {
      fprintf(fo,"ERROR: unable to execute crafty!\n");
      fclose(fo);
      }
    exit(1);
    }
  close(tp[0]);
  close(fp[1]);
  boardinit();

 
  write(to_crafty,"xboard\n",7);
  sleep(1);
  write(to_crafty,"protover 2\n",11);
  sleep(1);  
  write(to_crafty,"output long\n",12);
  sleep(1);  
  write(to_crafty,"level 0 5 0\n",12);
  
  
  i=0;
  j=0;
  k=0;

  
  fcntl(fileno(stdin),F_SETFL,O_NONBLOCK);
  fcntl(from_crafty,F_SETFL,O_NONBLOCK);
  
  
  side = 0;
  lock =0;
  postpone=0;


  n = (fd > from_crafty ? fd : from_crafty) + 1;

  while(1)
    {
    FD_ZERO(&fdread);
    FD_SET(fileno(stdin),&fdread);
    FD_SET(fd,&fdread);
    FD_SET(from_crafty,&fdread);

    tv.tv_sec=0;
    tv.tv_usec=100000;
    select(n,&fdread,NULL,NULL,&tv);
    do
      {
      if(FD_ISSET(from_crafty,&fdread))
        {
        while((readnum=read(from_crafty,&inbuffer[i],128))>0)
	  {
	  i+=readnum;
          while(excommand(command,inbuffer,i))
	    {
	    i-=strlen(command);
	    printf("ENGINE        :");
	    if(debug)
	      {
	      fprintf(fo,"ENGINE        :%3d:%3d:%s",strlen(command),i,command);
	      printf("%3d:%3d:    ",strlen(command),i);
	      }
	    printf("%s",command);
	    if(!strncmp(command,"move",4))
	      {
	      if(debug)
	        {
	        fprintf(fo,"ENGINE                :    Move ok\n");
	        printf("ENGINE                :    Move ok\n");	      
	        }
	      if(!lock)
	        {
	        movetoboard(&command[5]);
	        side = (side + 1)%3;
	        }
	      else
	        {
	        if(debug)
		  {
		  printf("Postponed\n");
		  fprintf(fo,"Postponed\n");
		  }
	        postpone =1;
	        strncpy(posmove,&command[5],9);
	        posmove[9]=0;
	        }
	      }
            }
	  fflush(stdout);
	  fflush(fo);
  	  FD_CLR(from_crafty,&fdread);
	  }
        }

      if(postpone && !lock)
        {
        if(debug)
	  {
	  printf("ENGINE                :    Moving\n");
	  fprintf(fo,"ENGINE                :    Moving\n");
	  }
        movetoboard(posmove);
        side = (side + 1)%3;
        postpone = 0;
        }

      if(FD_ISSET(fileno(stdin),&fdread))
        {
        while((readnum=read(fileno(stdin),&outbuffer[j],128))>0)
	  {
	  j+=readnum;
          while(excommand(command,outbuffer,j))
	    {
	    j-=strlen(command);
	    write(to_crafty,command,strlen(command));
	    }
	  }
        fflush(stdout);
        fflush(fo);
	FD_CLR(fileno(stdin),&fdread);
        }

      if(FD_ISSET(fd,&fdread))
        {
        readnum=read(fd,&boardbuffer[k],1024);
	if(readnum!=-1) 
          {
	  for(l=k;l<k+readnum;l++)
            boardbuffer[l] &= 0x7F;
          k+=readnum;
          while(excommand(command,boardbuffer,k))
	    {
	    k-=strlen(command);
	    if(debug)
	      {
	      fprintf(fo,"BOARD         :%3d:%3d:    %s",strlen(command),k,command);
    	      printf("BOARD         :");
	      printf("%3d:%3d:    ",strlen(command),k);
	      printf("%s",command);
	      }
	    if((strncmp(&command[4],"15",2))==0 && lock==1)
	      {
	      if(debug)
	        {
	        printf("Unlocked\n");
	        fprintf(fo,"Unlocked\n");
	        }
	      lock=0;
	      }
	    else if((strncmp(&command[4],"18",2))==0)
	      if(side==0)
	        {
	        if(debug)
	          {
		  printf("Ok\n");
		  fprintf(fo,"Ok\n");
   	          fflush(stdout);
		  }
	        movetocrafty(&command[12]);
	        lock = 1;
	        side = (side + 1)%3;
	        }
	      else
	        side = (side + 1)%3;
	    else if((strncmp(&command[4],"05",2))==0 || ((strncmp(&command[4],"0C",2))==0 && (strncmp(&command[20],"FF",2))==0 ))
	      {
	      side=0;
	      lock=0;
	      postpone=0;
	      write(to_crafty,"new\n",4);
	      if(debug)
	        {
	        fprintf(fo,"New Game.\n");
                }
	      printf("New Game.\n");
	      break;
	      }
	    else if((strncmp(&command[4],"15",2))==0 && side==0)
	      {
	      side = 1;
	      write(to_crafty,"force\n",6);
	      write(to_crafty,"go\n",3);
	      if(debug)
	        {
	        fprintf(fo,"Change side.\n");
                }	    
	      printf("Change side.\n");
	      }
	    else if((strncmp(&command[4],"16",2))==0 && side==1 && (!postpone))
	      {
	      write(to_crafty,"?\n",2);
	      if(debug)
	        {
	        fprintf(fo,"Move now.\n");
	        printf("Move now.\n");
                }	    
	      }
	    else if((strncmp(&command[4],"1E",2))==0)
	      {
	      write(to_crafty,"quit\n",5);
	      kill(crafty_pid,SIGTERM);
	      if(debug)
	        {
	        fprintf(fo,"Quit.\n");
	        fclose(fo);
                }	    
	      printf("Quit.\n");
	      exit(0);
	      }
	    }
	  }
        fflush(stdout);
        fflush(fo);
	FD_CLR(fd,&fdread);
        }
      } while(FD_ISSET(fileno(stdin),&fdread)||FD_ISSET(fd,&fdread)|FD_ISSET(from_crafty,&fdread));
    }
  }



void boardinit(void)
{
char move[28] = "0D020C0001C0800000040000FF\r";

 
/* 
  Open serial device for reading and writing and not as controlling tty
  because we don't want to get killed if linenoise sends CTRL-C.
*/
 fd = open(SERIALDEVICE, O_RDWR | O_NOCTTY );
 if (fd <0) {perror(SERIALDEVICE); exit(-1); }
 
 tcgetattr(fd,&oldtio); /* save current modem settings */
 
/* 
  Set bps rate and hardware flow control and 8n1 (8bit,no parity,1 stopbit).
  Also don't hangup automatically and ignore modem status.
  Finally enable receiving characters.
*/
 newtio.c_cflag = BAUDRATE | CRTSCTS | CS8 | CLOCAL | CREAD;
 
/*
 Ignore bytes with parity errors and make terminal raw and dumb.
*/
 newtio.c_iflag = IGNPAR;
 
/*
 Raw output.
*/
 newtio.c_oflag = 0;
 
/*
 Don't echo characters because if you connect to a host it or your
 modem will echo characters for you. Don't generate signals.
*/
 newtio.c_lflag = 0;
 
/* blocking read until 1 char arrives */
 newtio.c_cc[VMIN]=1;
 newtio.c_cc[VTIME]=0;

 
 

/* now clean the modem line and activate the settings for modem */
 tcflush(fd, TCIFLUSH);
 tcsetattr(fd,TCSANOW,&newtio);
 fcntl(fd,F_SETFL,O_NONBLOCK);

/* terminal settings done, now handle in/ouput */

   /* user input */
    /* close(1);  stdout not needed */
  write(fd,"}",1);
  sleep(1);
  write(fd,"0602010055AA\r",13);
  sleep(1);
  //  write(fd,"060211000195\r",13);
  //  write(fd,"0602110001B5\r",13);
  write(fd,"0602110001D5\r",13);

}

void movetoboard(char * move)
  {
  static char standmove[28] = "0D020C0001C0800000040000FF\r";
  static char promopiece[4]="NBRQ";
  char c;
  int i,j,k,promo;

  i=0;
  promo=-1;

  if(move[0]=='O' && move[2]=='O')
    if(move[3]=='-')
      {
      if(color==BLACK)
        movetoboard("e8c8");
      else
        movetoboard("e1c1");
      return;
      }
    else
      {
      if(color==BLACK)
        movetoboard("e8g8");
      else
	movetoboard("e1g1");
      return;
      }

  if(isupper(move[0])) i++;

  for (j=0;j<2;j++) 
    {
    standmove[20+j] = move[i+j] - (j%2 ? 1 : 'a' - '0');
    }

  if(move[i+2] == 'x') i++;

  for (j=2;j<4;j++) 
    {
    standmove[20+j] = move[i+j] - (j%2 ? 1 : 'a' - '0');
    }
  
  k=0;
  if(move[i+4]=='=')
    for(k=0;k<4;k++)
      if(move[i+5]==promopiece[k])
        break;

  if(k&1)
    standmove[23] += (standmove[23] < '2' ? 8 : 15);

  if(k&2)
    standmove[22] += (standmove[22] < '2' ? 8 : 15);

    sleep(1);
    write(fd,"!",1);
    do read(fd,&c,1);
    while ((c&0x7f) != '&');

    if(debug)
      {
      fprintf(fo,"TO_BOARD              :    %s\n",standmove);
      printf("TO_BOARD              :    %s\n",standmove);
      fflush(stdout);
      }
    write(fd,standmove,27);
  }

void movetocrafty(char * move)
  {
  int i;
  int sel;
  static char standmove[7]="e2e4 \n";
  static char promopiece[4]="NBRQ";
  sel = 0;

  for(i=0;i<4;i++)
    {
    standmove[i] = move[i] + (i%2 ? 1 : 'a' - '0');
    }

  if(move[4]>'7')
    {
    sel += ((move[4] - (move[4] > '9' ? 'A' - 10 : '0')) & 1) << 1;
    sel += ((move[5] - (move[5] > '9' ? 'A' - 10 : '0')) & 8) >> 3;
    standmove[4] = promopiece[sel];
    }

  printf("BOARD'S MOVE          :    %s",standmove);
  fflush(stdout);
  if(debug)
    fprintf(fo,"TO_ENGINE              :    %s",standmove);
    
  write(to_crafty,standmove,6);

  standmove[4] = ' ';
  }

int excommand(char *command,char * buffer,int n)
  {
  int i,j,found;

  found=FALSE;

  for(i=0;i<n;i++)
    {
    if(buffer[i]=='\n')
      {
      found=TRUE;
      break;
      }
    }

  if(found)
    {
    for(j=0;j<i+1;j++)
      command[j] = buffer[j] =='\r' ? ' ' : buffer[j];
    command[j]=0;

    for(j=0;j<n;j++)
      buffer[j]=buffer[i+j+1];
    }

  return found;
  }
